@@ -1,18 +1,21 @@
-from typing import List, Tuple, Optional
+import copy
+from typing import List, Tuple, Optional, Callable, Dict
 
 import torch
 import torch.nn.functional as f
+import torchmetrics
 from einops import rearrange
 from pytorch_lightning.core import LightningModule
 from torch import nn, Tensor
 import numpy as np
+from torchmetrics import Metric
 
 from lit_saint.config import SaintConfig
 from lit_saint.modules import SimpleMLP, RowColTransformer, SepMLP
 from lit_saint.augmentations import cutmix, mixup, get_random_index
 
 
-class SAINT(LightningModule):
+class Saint(LightningModule):
     """Contains the definition of the network and how to execute the pretraining tasks, and how perform
     the training, validation and test steps
 
@@ -20,8 +23,10 @@ class SAINT(LightningModule):
     :param continuous: List of indices with continuous columns
     :param dim_target: if categorical represent number of classes of the target otherwise is 1
     :param config: configuration of the model
-    :param pretraining: boolean flag, if True it will be executed pretraining task
-    :param mc_dropout: boolean flag, if true mc_dropout will be applied during prediction
+    :param metrics: Dictionary containing custom metrics to compute during training loop
+    :param metrics_single_class: boolean flag, if True the metrics are computed separately for each class
+    :param optimizer: custom optimizer to compute gradient of the network
+    :param loss_fn: custom loss function to be optimized
     """
     def __init__(
             self,
@@ -29,15 +34,19 @@ class SAINT(LightningModule):
             continuous: List[int],
             dim_target: int,
             config: SaintConfig,
-            pretraining: bool = False,
-            mc_dropout: bool = False
+            metrics: Dict[str, Metric] = None,
+            metrics_single_class: bool = True,
+            optimizer: Callable = torch.optim.Adam,
+            loss_fn: Callable = None,
     ):
         super().__init__()
         self.save_hyperparameters()
         assert all(map(lambda n: n > 0, categories)), 'number of each category must be positive'
-        self.mc_dropout = mc_dropout
+        self.loss_fn = loss_fn
+        self.optimizer = optimizer
+        self.mc_dropout = False
         self.config = config
-        self.pretraining = pretraining
+        self.pretraining = False
         self.dim_target = dim_target
         self.num_categories = len(categories)
         self.num_unique_categories = sum(categories)
@@ -47,6 +56,26 @@ class SAINT(LightningModule):
         cat_mask_offset = f.pad(torch.tensor(categories), (1, 0), value=0)
         self.cat_mask_offset = cat_mask_offset.cumsum(dim=-1)[:-1]
         self._define_network_components(categories)
+        self.metrics_single_class = metrics_single_class
+        self.metrics = metrics if metrics else {}
+        self.train_metrics = self._define_metrics(self.metrics, metrics_single_class)
+        self.val_metrics = self._define_metrics(self.metrics, metrics_single_class)
+
+    def _define_metrics(self, metrics: Dict[str, Metric], metrics_single_class: bool) -> Dict[str, Metric]:
+        """Define custom metrics computed during training loop
+
+        :param metrics: Custom metrics to compute
+        :param metrics_single_class: boolean flag, if True the metrics are computed separately for each class
+        """
+        metrics_step = {}
+        if metrics_single_class:
+            for key, value in metrics.items():
+                for i in range(self.dim_target):
+                    metrics_step[f"{key}_{i}"] = copy.deepcopy(value)
+        else:
+            for key, value in metrics.items():
+                metrics_step[key] = copy.deepcopy(value)
+        return metrics_step
 
     def _define_network_components(self, categories: List[int]) -> None:
         """Define the components for the neural network
@@ -58,18 +87,21 @@ class SAINT(LightningModule):
         self._define_mlp(categories)
         self._define_projection_head()
         self.mlpfory = SimpleMLP(self.config.network.embedding_size * self.num_columns,
-                                 self.config.network.internal_dimension_output_layer, self.dim_target,
-                                 dropout=self.config.network.ff_dropout)
+                                 self.config.train.internal_dimension_output_layer, self.dim_target,
+                                 dropout=self.config.train.mlpfory_dropout)
 
     def _define_transformer(self) -> None:
         """Instantiate the type of Transformed that will be used in SAINT"""
         self.transformer = RowColTransformer(
             dim=self.config.network.embedding_size,
             nfeats=self.num_categories + self.num_continuous,
-            depth=self.config.network.depth,
-            heads=self.config.network.heads,
-            ff_dropout=self.config.network.ff_dropout,
-            style=self.config.network.attention_type
+            depth=self.config.network.transformer.depth,
+            heads=self.config.network.transformer.heads,
+            dim_head=self.config.network.transformer.dim_head,
+            ff_dropout=self.config.network.transformer.dropout,
+            style=self.config.network.transformer.attention_type.value,
+            scale_dim_internal_col=self.config.network.transformer.scale_dim_internal_col,
+            scale_dim_internal_row=self.config.network.transformer.scale_dim_internal_row
         )
 
     def _define_embedding_layers(self) -> None:
@@ -77,27 +109,31 @@ class SAINT(LightningModule):
         # embed continuos variables using one different MLP for each column
         self.embedding_continuos = nn.ModuleList([SimpleMLP(1, self.config.network.internal_dimension_embed_continuous,
                                                             self.config.network.embedding_size,
-                                                            dropout=self.config.network.ff_dropout)
+                                                            dropout=self.config.network.dropout_embed_continuous)
                                                   for _ in range(self.num_continuous)])
         # embedding layer categorical columns
         self.embedding_categorical = nn.Embedding(self.num_unique_categories, self.config.network.embedding_size)
 
     def _define_mlp(self, categories: List[int]) -> None:
-        """Define MLP used for the inference"""
-        self.mlp1 = SepMLP(dim=self.config.network.embedding_size, dim_out_for_each_feat=categories)
+        """Define MLP used for the Denoising task"""
+        self.mlp1 = SepMLP(dim=self.config.network.embedding_size, dim_out_for_each_feat=categories,
+                           scale_dim_internal=self.config.pretrain.task.denoising.scale_dim_internal_sepmlp,
+                           dropout=self.config.pretrain.task.denoising.dropout)
         self.mlp2 = SepMLP(dim=self.config.network.embedding_size,
-                           dim_out_for_each_feat=list(np.ones(self.num_continuous).astype(int)))
+                           dim_out_for_each_feat=list(np.ones(self.num_continuous).astype(int)),
+                           scale_dim_internal=self.config.pretrain.task.denoising.scale_dim_internal_sepmlp,
+                           dropout=self.config.pretrain.task.denoising.dropout)
 
     def _define_projection_head(self) -> None:
         """Define projection heads for contrastive learning"""
         self.pt_mlp = SimpleMLP(self.config.network.embedding_size * self.num_columns,
                                 6 * self.config.network.embedding_size * self.num_columns // 5,
                                 self.config.network.embedding_size * self.num_columns // 2,
-                                dropout=self.config.network.ff_dropout)
+                                dropout=self.config.pretrain.task.contrastive.dropout)
         self.pt_mlp2 = SimpleMLP(self.config.network.embedding_size * self.num_columns,
                                  6 * self.config.network.embedding_size * self.num_columns // 5,
                                  self.config.network.embedding_size * self.num_columns // 2,
-                                 dropout=self.config.network.ff_dropout)
+                                 dropout=self.config.pretrain.task.contrastive.dropout)
 
     def _embed_data(self, x_categ: Tensor, x_cont: Tensor) -> Tuple[Tensor, Tensor]:
         """Converts categorical and continuous values in embeddings
@@ -112,12 +148,11 @@ class SAINT(LightningModule):
         x_cont_enc = torch.empty(n1, n2, self.config.network.embedding_size)
         for i in range(self.num_continuous):
             x_cont_enc[:, i, :] = self.embedding_continuos[i](x_cont[:, i])
-        x_cont_enc = x_cont_enc
 
         return x_categ_enc, x_cont_enc
 
     def _embeddings_contrastive(self, embed_categ: Tensor, embed_cont: Tensor, embed_categ_noised: Tensor,
-                                embed_cont_noised: Tensor, projhead_style: str = "different") -> Tuple[Tensor, Tensor]:
+                                embed_cont_noised: Tensor, projhead_style) -> Tuple[Tensor, Tensor]:
         """Given embeddings original and noised version of the categorical and continuous features are
         transformed using attention, normalized and eventually projected using MLP
 
@@ -133,10 +168,10 @@ class SAINT(LightningModule):
         embed_tranformed = (embed_tranformed / embed_tranformed.norm(dim=-1, keepdim=True)).flatten(1, 2)
         embed_transformed_noised = (embed_transformed_noised /
                                     embed_transformed_noised.norm(dim=-1, keepdim=True)).flatten(1, 2)
-        if projhead_style == 'different':
+        if projhead_style.value == 'different':
             embed_tranformed = self.pt_mlp(embed_tranformed)
             embed_transformed_noised = self.pt_mlp2(embed_transformed_noised)
-        elif projhead_style == 'same':
+        elif projhead_style.value == 'same':
             embed_tranformed = self.pt_mlp(embed_tranformed)
             embed_transformed_noised = self.pt_mlp(embed_transformed_noised)
         else:
@@ -203,7 +238,7 @@ class SAINT(LightningModule):
         :param embed_categ_noised: embeddings categorical features after the augmentation
         :param embed_cont_noised: embeddings continuous features after the augmentation
         """
-        if self.config.pretrain.task.contrastive:
+        if self.config.pretrain.task.contrastive.constrastive_type.value == 'standard':
             embed_tranformed, embed_transformed_noised = self._embeddings_contrastive(
                 embed_categ, embed_cont, embed_categ_noised, embed_cont_noised,
                 self.config.pretrain.task.contrastive.projhead_style)
@@ -216,20 +251,24 @@ class SAINT(LightningModule):
             loss_1 = f.cross_entropy(logits_1, targets)
             loss_2 = f.cross_entropy(logits_2, targets)
             return self.config.pretrain.task.contrastive.weight * (loss_1 + loss_2) / 2
-        elif self.config.pretrain.task.contrastive_sim:
+        elif self.config.pretrain.task.contrastive.constrastive_type.value == 'simsiam':
             # it apply the concept of simsiam we want that the embedding minimize the cosine similarity
             # the idea is that we want on the diagonal all 1, it means they are equal /because normalized)
             embed_tranformed, embed_transformed_noised = self._embeddings_contrastive(
-                embed_categ, embed_cont, embed_categ_noised, embed_cont_noised)
-            return - self.config.pretrain.task.contrastive_sim.weight * \
+                embed_categ, embed_cont, embed_categ_noised, embed_cont_noised,
+                self.config.pretrain.task.contrastive.projhead_style)
+            return - self.config.pretrain.task.contrastive.weight * \
                 f.cosine_similarity(embed_tranformed, embed_transformed_noised).add_(-1).sum()
 
     def configure_optimizers(self):
         if self.pretraining:
-            lr = self.config.network.learning_rate_pretraining
+            lr = self.config.pretrain.optimizer.learning_rate
+            other_params = self.config.pretrain.optimizer.other_params
         else:
-            lr = self.config.network.learning_rate
-        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
+            lr = self.config.train.optimizer.learning_rate
+            other_params = self.config.pretrain.optimizer.other_params
+
+        optimizer = self.optimizer(self.parameters(), lr=lr, **other_params)
         return optimizer
 
     def forward(self, x_categ: Tensor, x_cont: Tensor) -> Tensor:
@@ -255,34 +294,62 @@ class SAINT(LightningModule):
             loss += self._pretraining_denoising(x_categ, x_cont, embed_categ_noised, embed_cont_noised)
         return loss
 
-    def shared_step(self, batch: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
+    def shared_step(self, batch: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
         """It define the commons step executed during training, validation and test
 
         :param batch: Contains a batch of data
         """
         x_categ, x_cont, target = batch
         if self.pretraining:
-            return self.pretraining_step(x_categ, x_cont)
+            return self.pretraining_step(x_categ, x_cont), torch.Tensor(), torch.Tensor()
         else:
             y_pred = self(x_categ, x_cont)
-            if self.dim_target > 1:
-                loss = self._classification_loss(y_pred, target)
-                return loss
+            if self.loss_fn:
+                return self.loss_fn(y_pred, target), y_pred, target
             else:
-                return self._regression_loss(y_pred, target)
+                if self.dim_target > 1:
+                    loss = self._classification_loss(y_pred, target)
+                    return loss, y_pred, target
+                else:
+                    return self._regression_loss(y_pred, target), y_pred, target
 
     def training_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
-        loss = self.shared_step(batch)
+        loss, y_pred, target = self.shared_step(batch)
         self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
+        if not self.pretraining:
+            self.compute_metrics(metrics=self.train_metrics, y_pred=y_pred, target=target)
+            if len(self.train_metrics) > 0:
+                self.log("train_metrics", self.train_metrics, on_step=False, on_epoch=True, prog_bar=True, logger=True)
         return loss
 
+    def compute_metrics(self, metrics: Dict[str, Metric], y_pred: Tensor, target: Tensor) -> None:
+        """ Compute custom metrics during training loop
+
+        :param metrics: Metrics to compute
+        :param y_pred: predicted value
+        :param target: true value
+        """
+        pred = nn.Softmax(dim=-1)(y_pred) if self.dim_target > 1 else y_pred.detach()
+        for key, value in metrics.items():
+            if self.metrics_single_class:
+                index_class_metric = int(key.split("_")[-1])
+                target_single_class = target[target == index_class_metric]
+                if target_single_class.shape[0] > 0:
+                    value(pred[target == index_class_metric], target[target == index_class_metric])
+            else:
+                value(pred, target)
+
     def validation_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
-        loss = self.shared_step(batch)
+        loss, y_pred, target = self.shared_step(batch)
         self.log("val_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
+        if not self.pretraining:
+            self.compute_metrics(metrics=self.val_metrics, y_pred=y_pred, target=target)
+            if len(self.val_metrics) > 0:
+                self.log("val_metrics", self.val_metrics, on_step=False, on_epoch=True, prog_bar=True, logger=True)
         return loss
 
     def test_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
-        loss = self.shared_step(batch)
+        loss, _, _ = self.shared_step(batch)
         self.log("test_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
         return loss
 
@@ -316,3 +383,9 @@ class SAINT(LightningModule):
         :param target: Values to predict
         """
         return f.mse_loss(y_pred, target)
+
+    def set_pretraining(self, pretraining: bool):
+        self.pretraining = pretraining
+
+    def set_mcdropout(self, mc_dropout: bool):
+        self.mc_dropout = mc_dropout
